from collections.abc import Mapping
from typing import Any

from qcio import CalcSpec, CalcType

from qccodec.exceptions import EncoderError
from qccodec.models import NativeInput

SUPPORTED_CALCTYPES = {
    CalcType.energy,
    CalcType.gradient,
    CalcType.hessian,
    CalcType.optimization,
    CalcType.transition_state,
}
XYZ_FILENAME = "geometry.xyz"
PADDING = 20  # padding between keyword and value


def _validate_keywords(keywords: dict[str, Any]) -> None:
    """Validate keywords for ORCA encoder. Expecting all lowercase keys."""

    _NON_BLOCKS = {  # disallowed top-level blocks → where to set instead
        "coords": ".structure",
    }

    _NON_BLOCK_KEYWORDS = {  # disallowed keys inside allowed blocks → where to set instead
        "method": {"method": ".model.method", "runtyp": ".calctype"},
        "basis": {"basis": ".model.basis"},
    }

    # 1) Blocks that should not appear as top-level keywords
    for block, where in _NON_BLOCKS.items():
        if block in keywords:
            raise EncoderError(
                f"Block '{keywords[block]}' should not be set as a keyword. "
                f"Set its data on '{where}'."
            )

    # 2) Disallowed keys inside allowed blocks
    for block, disallowed in _NON_BLOCK_KEYWORDS.items():
        for variable, where in disallowed.items():
            if keywords.get(block, {}).get(variable) is not None:
                raise EncoderError(
                    f"Keyword '{variable}' in block '{block}' should not be set directly. It "
                    f"should be set at '{where}'",
                )


def _fmt(key: str, value: Any) -> Any:
    """Format a value for ORCA input."""
    keywords_needing_quotes = {"auxc", "auxj", "auxjk"}
    if key.casefold() in keywords_needing_quotes:
        return f'"{value}"'  # ORCA needs quotes for certain keywords
    if isinstance(value, bool):
        return str(value).lower()  # ORCA expects 'true'/'false' for booleans
    return value


def encode(inp_obj: CalcSpec) -> NativeInput:
    """Translate a CalcSpec into ORCA input files.

    Args:
        inp_obj: The qcio CalcSpec object for a computation.

    Returns:
        NativeInput with .input being an orca.inp file and .geometry an xyz file.

    Notes:
        - ORCA keywords are case-insensitive. This encoder will preserve the
            casing of keywords as provided in inp_obj.keywords.
        - ORCA requires passing `numgrad` for numerical gradients. To activate a
            numerical gradient for a single point or optimization pass
            `{"numgrad": "true"} in the keywords or `{"numgrad": {...}}` with a
            dictionary of numgrad block keywords. An empty dictionary will also work
            (equivalent to `{"numgrad": "true"}`).
    """

    # Handle ORCA's case-insensitive keywords by doing caseless lookups
    kw_lower = {k.casefold(): v for k, v in inp_obj.keywords.items()}
    _validate_keywords(kw_lower)

    # Collect lines for input file
    inp_lines = []

    # maxcore
    if "maxcore" in kw_lower:
        inp_lines.append(f"%maxcore {kw_lower['maxcore']}")

    # pal
    if "pal" in kw_lower:
        inp_lines.append(f"%pal nprocs {kw_lower['pal']} end")

    # Add an empty line if writing maxcore/pal
    if inp_lines:
        inp_lines.append("")

    # Method and Basis
    inp_lines.append(f"! {inp_obj.model.method} {inp_obj.model.basis}")

    # NumGrad. May be used for gradients or optimizations
    if "numgrad" in kw_lower:
        inp_lines.append(f"! numgrad")

    # Set ORCA runtyp based on calctype
    if inp_obj.calctype == CalcType.energy:
        runtyp = "energy"
    elif inp_obj.calctype == CalcType.gradient:
        runtyp = "engrad"
    elif inp_obj.calctype == CalcType.hessian:
        if "numfreq" in kw_lower:
            runtyp = "numfreq"
        else:
            runtyp = "freq"
    elif inp_obj.calctype == CalcType.optimization:
        runtyp = "opt"
    elif inp_obj.calctype == CalcType.transition_state:
        runtyp = "optts"

    inp_lines.append(f"! {runtyp}\n")

    # Input Blocks
    for block, kwargs in kw_lower.items():
        if isinstance(kwargs, Mapping):  # Skip non-block keywords
            inp_lines.append(f"%{block}")
            for key, value in kwargs.items():
                inp_lines.append(f"    {key:<{PADDING}} {_fmt(key, value)}")
            inp_lines.append("end")

    # Write a new line if there were any blocks
    if any(isinstance(v, Mapping) for v in kw_lower.values()):
        inp_lines.append("")

    # Structure
    inp_lines.append(
        f"* xyzfile {inp_obj.structure.charge} {inp_obj.structure.multiplicity} {XYZ_FILENAME}"
    )

    return NativeInput(
        input_file="\n".join(inp_lines) + "\n",
        geometry_file=inp_obj.structure.to_xyz(),
        geometry_filename=XYZ_FILENAME,
    )



import math
import re
from collections.abc import Generator
from enum import Enum
from pathlib import Path
from typing import Any

import numpy as np
from qcconst import constants
from qcio import (
    CalcSpec,
    CalcType,
    Provenance,
    Results,
    SinglePointData,
    Structure,
)

from qccodec.exceptions import ParserError

from ..registry import register
from .utils import re_finditer, re_search


class CrestFileType(str, Enum):
    """CREST filetypes.

    Maps file types to their names as written in the CREST output directory (except
    for STDOUT and DIRECTORY).
    """

    STDOUT = "stdout"
    DIRECTORY = "directory"
    NUMHESS = "numhess1"
    G98 = "g98.out"
    ENGRAD = "crest.engrad"
    OPTLOG = "crestopt.log"


def iter_files(
    stdout: str | None, directory: Path | str | None
) -> Generator[tuple[CrestFileType, str | bytes | Path], None, None]:
    """
    Iterate over the files in a CREST output directory.

    If stdout is provided, yields a tuple for it.

    If directory is provided, iterates over the directory to yield files according to
    program-specific logic.

    Args:
        stdout: The contents of the CREST stdout file.
        directory: The path to the directory containing the CREST output files.

    Yields:
        (FileType, contents) tuples for a program's output.
    """
    directory = Path(directory) if directory else None

    if stdout is not None:
        yield CrestFileType.STDOUT, stdout

    if directory is not None:
        # Check if the directory exists and is a directory
        if not directory.exists() or not directory.is_dir():
            raise ParserError(
                f"Directory {directory} does not exist or is not a directory."
            )
        yield CrestFileType.DIRECTORY, directory

        # Iterate over the files in the directory and yield their contents
        for filetype in CrestFileType:
            # Ignore STDOUT and DIRECTORY as they are handled above
            if filetype not in (CrestFileType.STDOUT, CrestFileType.DIRECTORY):
                # Check if the file exists in the directory
                file_path = directory / filetype.value
                if file_path.exists():
                    yield filetype, file_path.read_text()


@register(filetype=CrestFileType.STDOUT, target=("extras", "program_version"))
def parse_version(string: str) -> str:
    """Parse version string from CREST stdout.

    Matches format of 'crest --version' on command line.
    """
    regex = r"Version (\d+\.\d+\.\d+),"
    match = re_search(regex, string)
    return match.group(1)


@register(filetype=CrestFileType.DIRECTORY, calctypes=[CalcType.conformer_search])
def parse_conformers(
    directory: Path | str, stdout: str | None, input_data: CalcSpec
) -> dict[str, Any]:
    """Parse the conformers from the output directory of a CREST conformer search calculation.

    Args:
        stdout: The contents of the CREST stdout file (not used).
        directory: Path to the directory containing the CREST output files.
        input_data: The input object used for the calculation

    Returns:
        The parsed conformers and their energies as a dictionary.
    """
    directory = Path(directory)
    conformers = Structure.open_multi(
        directory / "crest_conformers.xyz",
        charge=input_data.structure.charge,
        multiplicity=input_data.structure.multiplicity,
    )

    # CREST places the energy as the only value in the comment line
    conf_energies = [
        float(conf.extras[Structure._xyz_comment_key][0]) for conf in conformers
    ]

    # Add identifiers to the conformers if topo is unchanged
    _add_identifiers_to_structures(conformers, input_data)

    return {
        "conformers": conformers,
        "conformer_energies": conf_energies,
    }


@register(filetype=CrestFileType.DIRECTORY, calctypes=[CalcType.conformer_search])
def parse_rotamers(
    directory: Path | str, stdout: str | None, input_data: CalcSpec
) -> dict[str, Any]:
    """Parse the rotamers from the output directory of a CREST conformer search calculation.

    Args:
        stdout: The contents of the CREST stdout file (not used).
        directory: Path to the directory containing the CREST output files.
        input_data: The input object used for the calculation

    Returns:
        The parsed rotamers and their energies as a dictionary.
    """
    directory = Path(directory)
    rotamers = Structure.open_multi(
        directory / "crest_rotamers.xyz",
        charge=input_data.structure.charge,
        multiplicity=input_data.structure.multiplicity,
    )

    # CREST places the energy as the only value in the comment line
    conf_energies = [
        float(conf.extras[Structure._xyz_comment_key][0]) for conf in rotamers
    ]

    # Add identifiers to the rotamers if topo is unchanged
    _add_identifiers_to_structures(rotamers, input_data)

    return {
        "rotamers": rotamers,
        "rotamer_energies": conf_energies,
    }


def _add_identifiers_to_structures(
    structures: list[Structure], input_data: CalcSpec
) -> None:
    """Add identifiers to a list of Structure objects if the 'topo' is unchanged."""
    if input_data.keywords.get("topo", True):
        ids = input_data.structure.identifiers.model_dump()
        for struct in structures:
            struct.add_identifiers(**ids)


@register(
    filetype=CrestFileType.ENGRAD,
    calctypes=[CalcType.energy, CalcType.gradient],
    target="energy",
)
def parse_energy(contents: str) -> float:
    """Parse the output of a CREST energy and gradient calculation.

    Args:
        contents: The text of the output file.

    Returns:
        The parsed energy as a float.
    """
    energy_regex = r"# Energy \( Eh \)\n#*\n\s*([-\d.]+)"
    return float(re_search(energy_regex, contents).group(1))


@register(
    filetype=CrestFileType.ENGRAD,
    calctypes=[CalcType.energy, CalcType.gradient],
    target="gradient",
)
def parse_gradient(contents: str) -> list[list[float]]:
    """Parse the output of a CREST energy and gradient calculation.

    Args:
        contents: The text of the output file.

    Returns:
        The parsed gradient as a Nx3 list of lists of floats.
    """
    gradient_regex = r"# Gradient \( Eh/a0 \)\n#\s*\n((?:\s*[-\d.]+\n)+)"
    vals = [float(x) for x in re_search(gradient_regex, contents).group(1).split()]
    # Group the values into chunks of 3 (for x, y, z).
    return [vals[i : i + 3] for i in range(0, len(vals), 3)]


@register(filetype=CrestFileType.STDOUT, calctypes=[CalcType.hessian], target="energy")
def parse_energy_numhess(contents: str) -> float:
    """Parse the initial singlepoint calculation energy from stdout.

    Args:
        contents: The text of the stdout file.

    Returns:
        The parsed energy.
    """
    energy_regex = r"Energy\s=\s*([-+]?\d+\.\d+)\s*Eh"
    return float(re_search(energy_regex, contents).group(1))


@register(
    filetype=CrestFileType.NUMHESS, calctypes=[CalcType.hessian], target="hessian"
)
def parse_numhess1(contents: str) -> list[list[float]]:
    """Parse the numerical Hessian from the CREST numhess1 file.

    Args:
        contents: The text of the numhess1 file.

    Returns:
        The parsed Hessian as a list of lists of floats.
    """
    float_regex = r"[-]?\d*\.\d+|\d+"
    numbers = [float(match.group()) for match in re_finditer(float_regex, contents)]
    sqrt_n = int(math.sqrt(len(numbers)))
    if sqrt_n * sqrt_n != len(numbers):
        raise ParserError(
            f"Expected a square matrix, but found {len(numbers)} elements."
        )
    # Reshape the flat list into a square matrix
    return [[numbers[i * sqrt_n + j] for j in range(sqrt_n)] for i in range(sqrt_n)]


@register(
    filetype=CrestFileType.G98, calctypes=[CalcType.hessian], target="freqs_wavenumber"
)
def parse_g98_freqs(contents: str) -> list[float]:
    """Parse the frequencies (wavenumbers) from G98 output text.

    Args:
        contents: The text of the g98 output file.

    Returns:
        A list of frequencies (wavenumbers) as floats.
    """
    regex = r"Frequencies\s+--\s+(?P<floats>(?:-?\d+\.\d+\s*)+)"
    # matches ['-335.2821                75.3406                87.4971\n ', ...]
    matches = re_finditer(regex, contents)
    return [float(freq) for match in matches for freq in match.group("floats").split()]


@register(
    filetype=CrestFileType.G98,
    calctypes=[CalcType.hessian],
    target="normal_modes_cartesian",
)
def parse_g98_normal_modes(contents: str) -> list[list[list[float]]]:
    """Parse the normal mode displacements from G98 output text.

    Args:
        contents: The text of the G98 output file.

    Returns:
        An (n_modes, n_atoms, 3) NumPy array representing the normal mode displacements
        with cartesian coordinates in Bohr.
    """
    block_re = re.compile(r"(Frequencies\s+--.*?)(?=Frequencies\s+--|$)", re.DOTALL)
    floats_re = re.compile(r"[-+]?\d*\.\d+")
    normal_modes_cartesian = []

    # Find all blocks of frequencies and displacements
    for block in block_re.findall(contents):
        # Determine the number of frequencies in this block from freqs line.
        n_freqs = len(block.splitlines()[0].split()[2:])
        # Match all floats in the block after "Atom AN"
        displacements = floats_re.findall(block.split("Atom AN")[1])
        # Initialize an empty list for each mode.
        mode_disp: list[list[list[float]]] = [[] for _ in range(n_freqs)]
        n_atoms = len(displacements) // (3 * n_freqs)
        for i in range(n_atoms):
            for j in range(n_freqs):
                index = i * (3 * n_freqs) + j * 3
                # Convert the displacement values to floats and scale to Bohr
                coords = [
                    float(val) * constants.ANGSTROM_TO_BOHR
                    for val in displacements[index : index + 3]
                ]
                mode_disp[j].append(coords)
        normal_modes_cartesian.extend(mode_disp)

    return normal_modes_cartesian


@register(
    filetype=CrestFileType.DIRECTORY,
    calctypes=[CalcType.optimization],
    target="trajectory",
)
def parse_trajectory(
    directory: Path | str,
    stdout: str,
    input_data: CalcSpec,
) -> list[Results]:
    """Parse the output directory of a CREST optimization calculation.

    Args:
        directory: Path to the directory containing the CREST output files.
        stdout: The stdout from CREST.
        input_data: The input object used for the calculation.

    Returns:
        The parsed optimization results as a list of Results objects.
    """
    # Read in the xyz file containing the trajectory
    directory = Path(directory)
    xyz_text = (directory / "crestopt.log").read_text()

    # Parse structures and energies from the xyz file
    structures = Structure.from_xyz_multi(
        xyz_text,
        charge=input_data.structure.charge,
        multiplicity=input_data.structure.multiplicity,
    )
    energies = [
        float(struct.extras[Structure._xyz_comment_key][1]) for struct in structures
    ]

    # Fake gradient for each step because CREST does not output it
    fake_gradient = np.zeros(len(input_data.structure.symbols) * 3)

    # Parse program version
    program_version = parse_version(stdout)

    # Create the optimization trajectory
    trajectory: list[Results] = [
        Results(
            input_data=CalcSpec(
                calctype=CalcType.gradient,
                structure=struct,
                model=input_data.model,
            ),
            success=True,
            data=SinglePointData(energy=energy, gradient=fake_gradient),
            provenance=Provenance(
                program="crest",
                program_version=program_version,
            ),
        )
        for struct, energy in zip(structures, energies)
    ]

    # Collect final gradient if calculation succeeded
    enegrad = directory / CrestFileType.ENGRAD.value
    if enegrad.exists():
        # Parse the energy and gradient from the file
        contents = enegrad.read_text()
        gradient = parse_gradient(contents)
        # Fill in final gradient
        trajectory[-1].data.gradient[:] = gradient

    else:
        # Calculation failed, so set the last .success = False
        final_po = trajectory[-1].model_dump()
        final_po["success"] = False
        trajectory[-1] = Results(**final_po)

    return trajectory



import pytest
from qcio import CalcSpec, CalcType, Model
from qcio.utils import water

from qccodec.encoders.orca import _validate_keywords, encode
from qccodec.exceptions import EncoderError


@pytest.mark.parametrize(
    "calctype, method, basis, keywords",
    [
        (CalcType.energy, "b3lyp", "def2-svp", {}),
        (CalcType.energy, "b3lyp", "def2-svp", {"maxcore": 500, "pal": 4}),
        (CalcType.energy, "b3lyp", "def2-svp", {"scf": {"convergence": "verytight"}}),
        (
            CalcType.energy,
            "revdsd-pbep86-d4/2021",
            "def2-svp",
            {"basis": {"auxc": "def2-svp/c"}},
        ),
        (CalcType.gradient, "b3lyp", "def2-svp", {}),
        (
            CalcType.gradient,
            "revdsd-pbep86-d4/2021",
            "def2-svp",
            {"basis": {"auxc": "def2-svp/c"}, "numgrad": True},
        ),
        (
            CalcType.gradient,
            "revdsd-pbep86-d4/2021",
            "def2-svp",
            {"basis": {"auxc": "def2-svp/c"}, "numgrad": {"accuracy": 6, "dx": 0.002}},
        ),
        (CalcType.hessian, "b3lyp", "def2-svp", {}),
        (
            CalcType.hessian,
            "revdsd-pbep86-d4/2021",
            "def2-svp",
            {"basis": {"auxc": "def2-svp/c"}, "freq": {"numfreq": True}},
        ),
        (CalcType.optimization, "b3lyp", "def2-svp", {"geom": {"maxiter": 30}}),
        (
            CalcType.optimization,
            "b3lyp",
            "def2-svp",
            {"geom": {"maxiter": 30}, "numgrad": True},
        ),
        (CalcType.transition_state, "b3lyp", "def2-svp", {"geom": {"calc_hess": True}}),
        (
            CalcType.transition_state,
            "revdsd-pbep86-d4/2021",
            "def2-svp",
            {
                "basis": {"auxc": "def2-svp/c"},
                "geom": {"calc_hess": True, "numhess": True},
            },
        ),
    ],
)
def test_write_input_files(
    calctype: CalcType, method: str, basis: str, keywords: dict[str, object]
):
    """Test write_input_files method."""
    inp_obj = CalcSpec(
        calctype=calctype,
        model=Model(method=method, basis=basis),
        structure=water,
        keywords=keywords,
    )
    native_input = encode(inp_obj)
    input_file = native_input.input_file

    # Check that all defined main / block keywords ended up in the file
    for keyword, value in keywords.items():
        assert keyword in input_file, f"{keyword} not in\n{input_file}"

        if isinstance(value, dict):
            for block_keyword in value:
                assert block_keyword in input_file, (
                    f"{block_keyword} not in\n{input_file}"
                )


def test_validate_keywords_raises_error_if_coords_block_in_keywords():
    """The 'coords' block should not be set directly as a keyword."""

    keywords = {"coords": {"units": "angstrom"}}
    with pytest.raises(EncoderError):
        _validate_keywords(keywords)


def test_validate_keywords_raises_error_if_method_block_contains_method_keyword():
    """The 'method' keyword should not be set inside the 'method' block."""
    keywords = {"method": {"method": "b3lyp"}}
    with pytest.raises(EncoderError):
        _validate_keywords(keywords)


def test_validate_keywords_raises_error_if_method_block_contains_runtyp_keyword():
    """The 'runtyp' keyword should not be set inside the 'method' block."""
    keywords = {"method": {"runtyp": "sp"}}
    with pytest.raises(EncoderError):
        _validate_keywords(keywords)


def test_validate_keywords_raises_error_if_basis_block_contains_basis_keyword():
    """The 'basis' keyword should not be set inside the 'basis' block."""
    keywords = {"basis": {"basis": "def2-svp"}}
    with pytest.raises(EncoderError):
        _validate_keywords(keywords)
